#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.hint module

Field title hint adapter implementation.
"""

from zope.interface import Interface
from zope.schema.interfaces import IField

from pyams_form.interfaces import IValue
from pyams_form.interfaces.form import IForm
from pyams_form.interfaces.widget import IWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@adapter_config(name='title',
                required=(Interface, IFormLayer, IForm, IField, IWidget),
                provides=IValue)
class FieldDescriptionAsHint:
    """Schema field description as widget's ```Title`` IValue adapter"""

    def __init__(self, context, request, form, field, widget):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.form = form
        self.field = field
        self.widget = widget

    def get(self):
        """Return the value"""
        return self.field.description if self.field.description else None



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.widget module

Base widget implementations.
"""

from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_form.browser.interfaces import IHTMLFormElement, IHTMLInputWidget, IHTMLSelectWidget, \
    IHTMLTextAreaWidget, IHTMLTextInputWidget
from pyams_form.interfaces import INPUT_MODE
from pyams_form.interfaces.widget import IFieldWidget


__docformat__ = 'restructuredtext'


class WidgetLayoutSupport:
    """Widget layout support mix-in class"""

    @staticmethod
    def wrap_css_class(klass, pattern='%(class)s'):
        """Return a list of css class names wrapped with given pattern"""
        if klass is not None and pattern is not None:
            return [pattern % {'class': k} for k in klass.split()]
        return []

    def get_css_class(self, klass=None, error=None, required=None,
                      class_pattern='%(class)s', error_pattern='%(class)s-error',
                      required_pattern='%(class)s-required'):
        # pylint: disable=too-many-arguments
        """Setup given css class (klass) with error and required postfix

        If no klass name is given the widget.wrapper class name/names get used.
        It is also possible if more then one (empty space separated) names
        are given as klass argument.

        This method can get used from your form or widget template or widget
        layout template without to re-implement the widget itself just because
        you a different CSS class concept.

        The following sample:

        <div tal:attributes="class python:widget.getCSSClass('foo bar')">
          label widget and error
        </div>

        will render a div tag if the widget field defines required=True:

        <div class="foo-error bar-error foo-required bar-required foo bar">
          label widget and error
        </div>

        And the following sample:

        <div tal:attributes="class python:widget.getCSSClass('row')">
          label widget and error
        </div>

        will render a div tag if the widget field defines required=True
        and an error occurs:

        <div class="row-error row-required row">
          label widget and error
        </div>

        Note; you need to define a globale widget property if you use
        python:widget (in your form template). And you need to use the
        view scope in your widget or layout templates.

        Note, you can set the pattern to None for skip error or required
        rendering. Or you can use a pattern like 'error' or 'required' if
        you like to skip postfixing your default css klass name for error or
        required rendering.

        """
        classes = []
        # setup class names
        if klass is not None:
            kls = klass
        else:
            kls = self.css  # pylint: disable=no-member

        # setup error class names
        if error is None:
            error = kls

        # setup required class names
        if required is None:
            required = kls

        # append error class names
        if self.error is not None:  # pylint: disable=no-member
            classes += self.wrap_css_class(error, error_pattern)
        # append required class names
        if self.required:  # pylint: disable=no-member
            classes += self.wrap_css_class(required, required_pattern)
        # append given class names
        classes += self.wrap_css_class(kls, class_pattern)
        # remove duplicated class names but keep order
        unique = []
        # pylint: disable=expression-not-assigned
        [unique.append(kls) for kls in classes if kls not in unique]
        return ' '.join(unique)


@implementer(IHTMLFormElement)
class HTMLFormElement(WidgetLayoutSupport):
    """HTML form element"""

    id = FieldProperty(IHTMLFormElement['id'])  # pylint: disable=invalid-name
    klass = FieldProperty(IHTMLFormElement['klass'])
    style = FieldProperty(IHTMLFormElement['style'])
    title = FieldProperty(IHTMLFormElement['title'])

    lang = FieldProperty(IHTMLFormElement['lang'])

    onclick = FieldProperty(IHTMLFormElement['onclick'])
    ondblclick = FieldProperty(IHTMLFormElement['ondblclick'])
    onmousedown = FieldProperty(IHTMLFormElement['onmousedown'])
    onmouseup = FieldProperty(IHTMLFormElement['onmouseup'])
    onmouseover = FieldProperty(IHTMLFormElement['onmouseover'])
    onmousemove = FieldProperty(IHTMLFormElement['onmousemove'])
    onmouseout = FieldProperty(IHTMLFormElement['onmouseout'])
    onkeypress = FieldProperty(IHTMLFormElement['onkeypress'])
    onkeydown = FieldProperty(IHTMLFormElement['onkeydown'])
    onkeyup = FieldProperty(IHTMLFormElement['onkeyup'])

    disabled = FieldProperty(IHTMLFormElement['disabled'])
    tabindex = FieldProperty(IHTMLFormElement['tabindex'])
    onfocus = FieldProperty(IHTMLFormElement['onfocus'])
    onblur = FieldProperty(IHTMLFormElement['onblur'])
    onchange = FieldProperty(IHTMLFormElement['onchange'])

    # layout support
    css = FieldProperty(IHTMLFormElement['css'])

    def add_class(self, klass):
        """See IHTMLFormElement"""
        if not self.klass:
            self.klass = str(klass)
        else:
            # make sure items are not repeated
            parts = self.klass.split() + [str(klass)]
            seen = {}
            unique = []
            for item in parts:
                if item in seen:
                    continue
                seen[item] = 1
                unique.append(item)
            self.klass = ' '.join(unique)

    def update(self):
        """See z3c.form.IWidget"""
        super().update()  # pylint: disable=no-member
        if self.mode == INPUT_MODE and self.required:  # pylint: disable=no-member
            self.add_class('required')


@implementer(IHTMLInputWidget)
class HTMLInputWidget(HTMLFormElement):
    """Base HTML input widget"""
    readonly = FieldProperty(IHTMLInputWidget['readonly'])
    alt = FieldProperty(IHTMLInputWidget['alt'])
    accesskey = FieldProperty(IHTMLInputWidget['accesskey'])
    onselect = FieldProperty(IHTMLInputWidget['onselect'])


@implementer(IHTMLTextInputWidget)
class HTMLTextInputWidget(HTMLInputWidget):
    """Base HTML text input widget"""
    size = FieldProperty(IHTMLTextInputWidget['size'])
    maxlength = FieldProperty(IHTMLTextInputWidget['maxlength'])
    placeholder = FieldProperty(IHTMLTextInputWidget['placeholder'])
    autocomplete = FieldProperty(IHTMLTextInputWidget['autocomplete'])
    autocapitalize = FieldProperty(IHTMLTextInputWidget['autocapitalize'])


@implementer(IHTMLTextAreaWidget)
class HTMLTextAreaWidget(HTMLFormElement):
    """Base HTML textarea widget"""
    rows = FieldProperty(IHTMLTextAreaWidget['rows'])
    cols = FieldProperty(IHTMLTextAreaWidget['cols'])
    readonly = FieldProperty(IHTMLTextAreaWidget['readonly'])
    accesskey = FieldProperty(IHTMLTextAreaWidget['accesskey'])
    onselect = FieldProperty(IHTMLTextAreaWidget['onselect'])


@implementer(IHTMLSelectWidget)
class HTMLSelectWidget(HTMLFormElement):
    """Base HTML select widget"""
    multiple = FieldProperty(IHTMLSelectWidget['multiple'])
    size = FieldProperty(IHTMLSelectWidget['size'])


def add_field_class(widget):
    """Add a class to the widget that is based on the field type name.

    If the widget does not have field, then nothing is done.
    """
    if IFieldWidget.providedBy(widget):
        klass = str(widget.field.__class__.__name__.lower() + '-field')
        widget.add_class(klass)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.ajex module

This module is declaring forms which can be used in an AJAX context, where
forms are submitted using an XMLHTTPRequest and where response is made of
JSON messages.
"""

import logging

import venusian
from pyramid.config import ConfigurationError
from pyramid.events import subscriber
from pyramid.interfaces import IRequest
from zope.component import queryMultiAdapter
from zope.interface import Interface, alsoProvides, implementer

from pyams_form.events import FormCreatedEvent
from pyams_form.interfaces import DISPLAY_MODE
from pyams_form.interfaces.error import IAJAXErrorsRenderer, IErrorViewSnippet, IMultipleErrors
from pyams_form.interfaces.form import IAJAXForm, IAJAXFormRenderer, IAddForm, IForm, \
    IFormCreatedEvent
from pyams_layer.interfaces import IFormLayer
from pyams_pagelet.interfaces import IPagelet
from pyams_pagelet.pagelet import Pagelet
from pyams_utils.adapter import adapter_config
from pyams_utils.dict import merge_dict
from pyams_utils.list import boolean_iter


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports

LOGGER = logging.getLogger('PyAMS (form)')


@implementer(IAJAXForm)
class AJAXForm:
    """A form mix-in class used to support form's submit through AJAX

    You don't have to inherit from AJAXForm or one of it's subclasses; they are used
    automatically when using the "ajax_form_config" class decorator.
    """

    def __call__(self):
        registry = self.request.registry  # pylint: disable=no-member
        registry.notify(FormCreatedEvent(self))
        self.update()  # pylint: disable=no-member
        has_errors, errors = boolean_iter(self.get_errors())  # pylint: disable=no-member
        if has_errors:
            return self.get_ajax_errors(errors)
        return self.get_ajax_output(self.finished_state.get('changes'))  # pylint: disable=no-member

    def get_ajax_errors(self, errors):
        """Get AJAX errors"""
        request = self.request  # pylint: disable=no-member
        registry = request.registry
        # pylint: disable=no-member
        renderer = registry.getMultiAdapter((self.context, request, self),
                                            IAJAXErrorsRenderer)
        return renderer.render(errors)

    def get_ajax_output(self, changes):
        """Get AJAX status and changes output in JSON"""
        raise NotImplementedError


@subscriber(IFormCreatedEvent, context_selector=IAJAXForm)
def handle_new_ajax_form(event):
    """Handle new AJAX form"""
    form = event.object
    impl = getattr(form, '__ajax_interfaces__', None)
    if impl is not None:
        alsoProvides(form, *impl)


class AJAXAddForm(AJAXForm):
    """AJAX add form mix-in class"""

    no_changes_message = _("No data was created.")

    def get_ajax_output(self, changes):
        request = self.request  # pylint: disable=no-member
        # pylint: disable=no-member
        renderer = None
        if 'action' in self.finished_state:
            name = self.finished_state['action'].field.getName()
            renderer = queryMultiAdapter((self.context, request, self),
                                         IAJAXFormRenderer, name=name)
        if renderer is None:
            renderer = queryMultiAdapter((self.context, request, self),
                                         IAJAXFormRenderer)
        if renderer is not None:
            result = renderer.render(changes)
            if result:
                return result
        if changes is None:
            return {
                'status': 'info',
                'message': request.localizer.translate(self.no_changes_message)
            }
        return {
            'status': 'reload'
        }


class AJAXEditForm(AJAXForm):
    """AJAX edit form mix-in class"""

    def get_ajax_output(self, changes):
        request = self.request  # pylint: disable=no-member
        translate = request.localizer.translate
        # pylint: disable=no-member
        result = {}
        renderer = None
        if 'action' in self.finished_state:
            name = self.finished_state['action'].field.getName()
            renderer = queryMultiAdapter((self.context, request, self),
                                         IAJAXFormRenderer, name=name)
        if renderer is None:
            renderer = queryMultiAdapter((self.context, request, self),
                                         IAJAXFormRenderer)
        if renderer is not None:
            result = renderer.render(changes) or {}
        status = ''
        message = ''
        if result:
            status = result.get('status')
            message = result.get('message')
        if not changes:
            if not status:
                result['status'] = 'info'
            if not message:
                result['message'] = translate(self.no_changes_message)
        else:
            if not status:
                result['status'] = status = 'success'
            if not message:
                if status == 'success':
                    result['message'] = translate(self.success_message)
        return result


@adapter_config(required=(None, IFormLayer, IAJAXForm),
                provides=IAJAXFormRenderer)
class AJAXFormRenderer:
    """Render an AJAX form's response into JSON format"""

    def __init__(self, context, request, form):
        self.context = context
        self.request = request
        self.form = form

    def render(self, changes):
        """Render form status and changes in JSON format"""
        registry = self.request.registry
        result = {}
        renderer = registry.queryMultiAdapter((self.context, self.request, self.form),
                                              IAJAXFormRenderer,
                                              name='main')
        if renderer is not None:
            result = renderer.render(changes)
        for form in self.form.get_forms(include_self=False):
            if form.mode == DISPLAY_MODE:
                continue
            renderer = registry.queryMultiAdapter((form.context, form.request, form),
                                                  IAJAXFormRenderer)
            if renderer is not None:
                form_output = renderer.render(changes)
                if form_output:
                    merge_dict(form_output, result)
        return result


@adapter_config(required=(Interface, IFormLayer, IForm),
                provides=IAJAXErrorsRenderer)
class AJAXErrorRenderer:
    """Default AJAX error renderer

    This renderer is based on status messages as handled by MyAMS framework.
    """

    def __init__(self, context, request, form):
        self.context = context
        self.request = request
        self.form = form

    @classmethod
    def get_widget_error(cls, error, status, translate):
        """Get widget for a given error"""
        message = translate(error.message)
        if hasattr(error, 'widget'):
            widget = error.widget
            if widget is not None:
                status.setdefault('widgets', []).append({
                    'id': widget.id,
                    'name': widget.name,
                    'label': translate(widget.label),
                    'message': message
                })
            else:
                error_message = {'message': message}
                if error_message not in status.get('messages', ()):
                    status.setdefault('messages', []).append(error_message)
        else:
            if message not in status.get('messages', ()):
                status.setdefault('messages', []).append(message)

    def render(self, errors):
        """Return status as JSON message"""
        registry = self.request.registry
        translate = self.request.localizer.translate
        result = {
            'status': 'error',
            'error_message': translate(self.form.status)
        }
        for error in errors:
            if isinstance(error, Exception):
                error = registry.getMultiAdapter(
                    (error, self.request, None, None, self, self.request),
                    IErrorViewSnippet)
            error.update()
            if IMultipleErrors.providedBy(error.error):
                for inner_error in error.error.errors:
                    self.get_widget_error(inner_error, result, translate)
            else:
                self.get_widget_error(error, result, translate)
        return result


class ajax_form_config:  # pylint: disable=invalid-name
    """Class decorator for form configuration with AJAX support

    When decorating a form class, this decorator create two new classes:
     - a first class, registered as a view and a pagelet, chich will provide classic HTML form
       rendering
     - a second class, inheriting from AJAX forms, which will handle AJAX queries executed when
       submitting the form; this class is registered as a Pyramid's views using a JSON renderer.

    Decorator arguments are all optional:
    :param name: AJAX view name, as used into browser's URL
    :param context (or *for_*): views context type
    :param layer: (or *request_type*): request layer for which the view is registered
    :param permission: permission required to call the base view
    :param ajax_name: name of the AJAX submit view; if undefined, this AJAX name will be the base
        view name, with ".html" replaced by ".json"
    :param ajax_base: base class for newly created AJAX forms; if not set, inherits from
        :py:class:`AJAXEditForm`of :py:classl`AJAXAddForm`, based on interface implemented by
        the view class
    :param ajax_implements: a list of marker interfaces implemented by the new class; these
        interfaces can be used, for example, to declare custom adapters based on these interfaces
    :param ajax_method: HTTP method name; if not defined, new AJAX
        view is restricted to *POST* requests; use "method=None" to be able to use any method
    :param ajax_permission: permission required to call the AJAX view; if not defined, AJAX view
        permission will be based on base view 'edit_permission', if any
    :param ajax_renderer: name of Pyramid renderer used to return AJAX view output; defaults to
        *json*
    :param ajax_xhr: Pyramid's view *xhr* predicate, *True* by default.

    All other view predicates are also available when using this decorator.
    """

    venusian = venusian

    def __init__(self, **settings):  # pylint: disable=too-many-branches
        if 'name' not in settings:
            raise ConfigurationError("Missing 'name' argument for form definition")
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings.pop('for_')
        if 'layer' in settings:
            settings['request_type'] = settings.pop('layer')
        # get AJAX settings
        settings['ajax'] = ajax_settings = {}
        if 'context' in settings:
            ajax_settings['context'] = settings.get('context')
        if 'ajax_name' in settings:
            ajax_settings['name'] = settings.pop('ajax_name')
        else:
            ajax_settings['name'] = settings['name'].replace('.html', '.json')
        if 'ajax_base' in settings:
            ajax_settings['base'] = settings.pop('ajax_base')
        if 'ajax_implements' in settings:
            ajax_settings['implements'] = settings.pop('ajax_implements')
        if 'ajax_method' in settings:
            method = settings.pop('ajax_method')
            if method is not None:
                ajax_settings['request_method'] = method
        else:
            ajax_settings['request_method'] = 'POST'
        if 'ajax_permission' in settings:
            ajax_settings['permission'] = settings.pop('ajax_permission')
        if 'ajax_renderer' in settings:
            renderer = settings.pop('ajax_renderer')
            if renderer is not None:
                ajax_settings['renderer'] = renderer
        else:
            ajax_settings['renderer'] = 'json'
        if 'ajax_xhr' in settings:
            xhr = settings.pop('ajax_xhr')
            if xhr is not None:
                ajax_settings['xhr'] = xhr
        else:
            ajax_settings['xhr'] = True
        if 'ajax_require_csrf' in settings:
            csrf = settings.pop('ajax_require_csrf')
            if csrf is not None:
                ajax_settings['require_csrf'] = csrf
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        form_settings = self.__dict__.copy()
        depth = form_settings.pop('_depth', 0)
        ajax_settings = form_settings.pop('ajax')

        def callback(context, name, obj):  # pylint: disable=unused-argument
            """Venusian decorator callback"""

            config = context.config.with_package(info.module)  # pylint: disable=no-member

            # Standard pagelet view settings
            form_cdict = {
                '__module__': obj.__module__,
                '__name__': form_settings.get('name'),
                'permission': form_settings.get('permission'),
                'ajax_form_handler': ajax_settings.get('name')
            }

            # AJAX view settings
            ajax_cdict = {
                '__module__': obj.__module__,
                '__name__': ajax_settings.get('name')
            }
            permission = ajax_settings.get('permission') or getattr(obj, '_edit_permission')
            if permission is not None:
                ajax_cdict['permission'] = permission
            base = ajax_settings.get('base')
            if base is None:
                if IAddForm.implementedBy(obj):  # pylint: disable=no-value-for-parameter
                    base = AJAXAddForm
                else:
                    base = AJAXEditForm
            if 'implements' in ajax_settings:
                impl = ajax_settings.pop('implements')
                if not isinstance(impl, (list, tuple, set)):
                    impl = (impl,)
                # classImplements(ajax_class, *impl)
                obj.__ajax_interfaces__ = impl

            form_class = type(obj.__name__, (obj, Pagelet), form_cdict)
            LOGGER.debug('Registering pagelet view "{}" for {} ({})'.format(
                form_settings.get('name'),
                str(form_settings.get('context', Interface)),
                str(form_class)))
            registry = form_settings.get('registry')
            if registry is None:
                config = context.config.with_package(info.module)  # pylint: disable=no-member
                registry = config.registry
            registry.registerAdapter(form_class,
                                     (form_settings.get('context', Interface),
                                      form_settings.get('request_type', IRequest)),
                                     IPagelet, form_settings.get('name'))
            # remove testing settings properties
            form_settings.pop('registry', None)
            form_settings.pop('venusian', None)
            config.add_view(view=form_class, **form_settings)

            ajax_class = type('AJAX' + obj.__name__, (base, obj), ajax_cdict)
            LOGGER.debug('Registering AJAX view "{0}" for {1} ({2})'.format(
                ajax_settings.get('name'),
                str(ajax_settings.get('context', Interface)),
                str(ajax_class)))
            # remove testing settings properties
            ajax_settings.pop('registry', None)
            ajax_settings.pop('venusian', None)
            config.add_view(view=ajax_class, **ajax_settings)

        info = self.venusian.attach(wrapped, callback, category='pyramid',
                                    depth=depth + 1)
        if info.scope == 'class':  # pylint: disable=no-member
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if form_settings.get('attr') is None:
                form_settings['attr'] = wrapped.__name__
            if ajax_settings.get('attr') is None:
                ajax_settings['attr'] = wrapped.__name__
        form_settings['_info'] = info.codeinfo  # pylint: disable=no-member
        ajax_settings['_info'] = info.codeinfo  # pylint: disable=no-member

        return wrapped



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.contentprovider module

Base form content providers.
"""

from zope.contentprovider.interfaces import IContentProvider
from zope.interface import Interface, Invalid, implementer
from zope.location import locate

from pyams_form.error import MultipleErrors
from pyams_form.field import FieldWidgets
from pyams_form.interfaces import DISPLAY_MODE, IContentProviders, IDataConverter, \
    IValidator
from pyams_form.interfaces.error import IErrorViewSnippet
from pyams_form.interfaces.form import IFieldsAndContentProvidersForm
from pyams_form.interfaces.widget import IWidgets
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import NO_VALUE


__docformat__ = 'restructuredtext'


class BaseProvider:
    """Base content provider"""

    __slots__ = ('position',)


_LOOKUP = BaseProvider()


@implementer(IContentProviders)
class ContentProviders(dict):
    """Content providers mapping"""

    def __init__(self, names=None):
        super().__init__()
        if names is not None:
            for position, name in enumerate(names):  # pylint: disable=unused-variable
                self[name] = _LOOKUP

    def __setitem__(self, key, value):
        factory = ContentProviderFactory(factory=value, name=key)
        super().__setitem__(key, factory)


class ContentProviderFactory:
    """Content provider factory"""

    def __init__(self, factory, name):
        self.factory = factory
        self.name = name
        self.position = getattr(factory, 'position', None)

    def __call__(self, manager):
        if self.factory != _LOOKUP:
            content_provider = self.factory(manager.content, manager.request, manager.form)
        else:
            registry = manager.request.registry
            content_provider = registry.getMultiAdapter((manager.content, manager.request,
                                                         manager.form),
                                                        IContentProvider, self.name)
        return content_provider


@adapter_config(required=(IFieldsAndContentProvidersForm, IFormLayer, Interface),
                provides=IWidgets)
class FieldWidgetsAndProviders(FieldWidgets):
    """Field widgets and providers adapter"""

    def update(self):
        super().update()
        unique_ordered_keys = list(self.keys())
        data = {}
        data.update(self)
        for name in self.form.content_providers:
            factory = self.form.content_providers[name]
            if factory.position is None:
                raise ValueError("Position of the following content provider should be an "
                                 "integer: '%s'." % name)
            content_provider = factory(self)
            short_name = name
            content_provider.update()
            unique_ordered_keys.insert(factory.position, short_name)
            data[short_name] = content_provider
            locate(content_provider, self, short_name)
        self.create_according_to_list(data, unique_ordered_keys)

    def extract(self):
        """See interfaces.IWidgets"""
        data = {}
        errors = ()
        registry = self.request.registry
        for name, widget in self.items():
            if IContentProvider.providedBy(widget):
                continue
            if widget.mode == DISPLAY_MODE:
                continue
            value = widget.field.missing_value
            try:
                widget.set_errors = self.set_errors
                raw = widget.extract()
                if raw is not NO_VALUE:
                    # pylint: disable=assignment-from-no-return
                    value = IDataConverter(widget).to_field_value(raw)
                registry.getMultiAdapter((self.content, self.request, self.form,
                                          getattr(widget, 'field', None), widget),
                                         IValidator).validate(value)
            except (Invalid, ValueError, MultipleErrors) as error:
                view = registry.getMultiAdapter((error, self.request, widget, widget.field,
                                                 self.form, self.content),
                                                IErrorViewSnippet)
                view.update()
                if self.set_errors:
                    widget.error = view
                errors += (view,)
            else:
                name = widget.__name__
                data[name] = value
        for error in self.validate(data):
            view = registry.getMultiAdapter((error, self.request, None, None,
                                             self.form, self.content),
                                            IErrorViewSnippet)
            view.update()
            errors += (view,)
        if self.set_errors:
            self.errors = errors
        return data, errors



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.events module

Data extraction event.
"""

from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent

from pyams_form.interfaces.form import IDataExtractedEvent, IFormCreatedEvent


__docformat__ = 'restructuredtext'


@implementer(IFormCreatedEvent)
class FormCreatedEvent(ObjectCreatedEvent):
    """Form created event"""

    def __init__(self, form):
        super().__init__(form)
        self.request = form.request


@implementer(IDataExtractedEvent)
class DataExtractedEvent:
    """Data extracted event"""

    def __init__(self, data, errors, form):
        self.data = data
        self.errors = errors
        self.form = form



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.widget module

Base widget implementation.
"""

from pyramid_chameleon.interfaces import IChameleonTranslate
from zope.interface import Invalid, alsoProvides, implementer, implementer_only
from zope.location import Location
from zope.schema import ValidationError
from zope.schema.fieldproperty import FieldProperty
from zope.schema.interfaces import IMinMaxLen, ITitledTokenizedTerm

from pyams_form.interfaces import IDataConverter, INPUT_MODE, ITerms, \
    IValidator, IValue, IWidgetLayoutTemplate
from pyams_form.interfaces.error import IErrorViewSnippet
from pyams_form.interfaces.form import IContextAware, IFormAware
from pyams_form.interfaces.widget import IAfterWidgetUpdateEvent, IFieldWidget, IMultiWidget, \
    ISequenceWidget, IWidget, IWidgetEvent
from pyams_form.template import get_widget_layout, get_widget_template
from pyams_form.util import sorted_none
from pyams_form.value import ComputedValueCreator, StaticValueCreator
from pyams_utils.interfaces.form import IDataManager, NO_VALUE
from pyams_utils.registry import query_utility


__docformat__ = 'restructuredtext'


PLACEHOLDER = object()

# pylint: disable=invalid-name
StaticWidgetAttribute = StaticValueCreator(
    discriminators=('context', 'request', 'view', 'field', 'widget')
)

# pylint: disable=invalid-name
ComputedWidgetAttribute = ComputedValueCreator(
    discriminators=('context', 'request', 'view', 'field', 'widget')
)


def apply_value_to_widget(parent, widget, value):
    """Apply given value to widget"""
    if value is not NO_VALUE:
        request = parent.request
        registry = request.registry
        try:
            # convert widget value to field value
            converter = IDataConverter(widget)
            # pylint: disable=assignment-from-no-return
            fvalue = converter.to_field_value(value)
            # validate field value
            registry.getMultiAdapter((parent.context, request, parent.form,
                                      getattr(widget, 'field', None), widget),
                                     IValidator).validate(fvalue)
            # convert field value back to widget value
            # that will probably format the value too
            # pylint: disable=assignment-from-no-return
            widget.value = converter.to_widget_value(fvalue)
        except (ValidationError, ValueError) as error:
            # on exception, setup the widget error message
            view = registry.getMultiAdapter((error, request, widget,
                                             widget.field, parent.form, parent.context),
                                            IErrorViewSnippet)
            view.update()
            widget.error = view
            # set the wrong value as value despite it's wrong
            # we want to re-show wrong values
            widget.value = value


@implementer(IWidget)
class Widget(Location):
    """Widget base class."""

    # widget specific attributes
    name = FieldProperty(IWidget['name'])
    label = FieldProperty(IWidget['label'])
    mode = FieldProperty(IWidget['mode'])
    required = FieldProperty(IWidget['required'])
    error = FieldProperty(IWidget['error'])
    value = FieldProperty(IWidget['value'])

    ignore_request = FieldProperty(IWidget['ignore_request'])
    ignore_required_on_validation = FieldProperty(IWidget['ignore_required_on_validation'])
    set_errors = FieldProperty(IWidget['set_errors'])
    show_default = FieldProperty(IWidget['show_default'])

    # rendering attributes
    layout = get_widget_layout()
    template = get_widget_template()

    # The following attributes are for convenience. They are declared in
    # extensions to the simple widget.

    # See ``interfaces.IContextAware``
    context = None
    ignore_context = False
    # See ``interfaces.IFormAware``
    form = None
    # See ``interfaces.IFieldWidget``
    field = None

    # Internal attributes
    _adapter_value_attributes = ('label', 'name', 'required', 'title')

    def __init__(self, request):
        self.request = request

    def update(self):
        """See z3c.form.interfaces.IWidget."""
        # Step 1: Determine the value.
        value = NO_VALUE
        look_for_default = False
        registry = self.request.registry
        # Step 1.1: If possible, get a value from the request
        if not self.ignore_request:
            # at this turn we do not need errors to be set on widgets
            # errors will be set when extract gets called from form.extractData
            self.set_errors = False
            widget_value = self.extract()
            if widget_value is not NO_VALUE:
                # Once we found the value in the request, it takes precendence
                # over everything and nothing else has to be done.
                self.value = widget_value
                value = PLACEHOLDER
        # Step 1.2: If we have a widget with a field and we have no value yet,
        #           we have some more possible locations to get the value
        if (IFieldWidget.providedBy(self) and
                value is NO_VALUE and
                value is not PLACEHOLDER):
            # Step 1.2.1: If the widget knows about its context and the
            #             context is to be used to extract a value, get
            #             it now via a data manager.
            if IContextAware.providedBy(self) and not self.ignore_context:
                value = registry.getMultiAdapter((self.context, self.field), IDataManager).query()
            # Step 1.2.2: If we still do not have a value, we can always use
            #             the default value of the field, if set
            # NOTE: It should check field.default is not missing_value, but
            # that requires fixing zope.schema first
            # We get a clone of the field with the context binded
            field = self.field.bind(self.context)

            if value is field.missing_value or value is NO_VALUE:
                default_value = field.default
                if default_value is not None and self.show_default:
                    value = field.default
                    look_for_default = True

        # Step 1.3: If we still have not found a value, then we try to get it
        #           from an attribute value
        if (value is NO_VALUE or look_for_default) and self.show_default:
            adapter = registry.queryMultiAdapter((self.context, self.request, self.form,
                                                  self.field, self),
                                                 IValue, name='default')
            if adapter:
                value = adapter.get()
        # Step 1.4: Convert the value to one that the widget can understand
        if value not in (NO_VALUE, PLACEHOLDER):
            converter = IDataConverter(self)
            # pylint: disable=assignment-from-no-return
            self.value = converter.to_widget_value(value)
        # Step 2: Update selected attributes
        for attr_name in self._adapter_value_attributes:
            # only allow to set values for known attributes
            if hasattr(self, attr_name):
                value = registry.queryMultiAdapter((self.context, self.request, self.form,
                                                    self.field, self),
                                                   IValue, name=attr_name)
                if value is not None:
                    setattr(self, attr_name, value.get())

    def extract(self, default=NO_VALUE):
        """See z3c.form.interfaces.IWidget."""
        return self.request.params.get(self.name, default)

    render = get_widget_template()

    def json_data(self):
        """Get widget data in JSON format"""
        return {
            'mode': self.mode,
            'error': self.error.message if self.error else '',
            'value': self.value,
            'required': self.required,
            'name': self.name,
            'id': getattr(self, 'id', ''),
            'type': 'text',
            'label': self.label or ''
        }

    def __call__(self, **kwargs):
        """Get and return layout template which is calling widget/render"""
        layout = self.layout
        if layout is None:
            registry = self.request.registry
            layout = registry.getMultiAdapter((self.context, self.request, self.form,
                                               self.field, self),
                                              IWidgetLayoutTemplate, name=self.mode)
        cdict = {
            'context': self.context,
            'request': self.request,
            'view': self,
            'translate': query_utility(IChameleonTranslate)
        }
        cdict.update(kwargs)
        return layout(**cdict)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.name)


@implementer(ISequenceWidget)
class SequenceWidget(Widget):
    """Term based sequence widget base.

    The sequence widget is used for select items from a sequence. Don't get
    confused, this widget does support to choose one or more values from a
    sequence. The word sequence is not used for the schema field, it's used
    for the values where this widget can choose from.

    This widget base class is used for build single or sequence values based
    on a sequence which is in most use case a collection. e.g.
    IList of IChoice for sequence values or IChoice for single values.

    See also the MultiWidget for build sequence values based on none collection
    based values. e.g. IList of ITextLine
    """

    value = ()
    terms = None
    separator = ','

    no_value_token = '--NOVALUE--'

    @property
    def display_value(self):
        """Widget display value"""
        value = []
        for token in self.value:
            # Ignore no value entries. They are in the request only.
            if token == self.no_value_token:
                continue
            try:
                term = self.terms.getTermByToken(token)
            except LookupError:
                # silently ignore missing tokens, because INPUT_MODE and
                # HIDDEN_MODE does that too
                continue
            if ITitledTokenizedTerm.providedBy(term):
                translate = self.request.localizer.translate
                value.append(translate(term.title))
            else:
                value.append(term.value)
        return value

    def update_terms(self):
        """Get terms from widget context"""
        if self.terms is None:
            registry = self.request.registry
            self.terms = registry.getMultiAdapter((self.context, self.request, self.form,
                                                   self.field, self),
                                                  ITerms)
        return self.terms

    def update(self):
        """See pyams_form.interfaces.widget.IWidget."""
        # Create terms first, since we need them for the generic update.
        self.update_terms()
        super().update()

    def extract(self, default=NO_VALUE):
        """See pyams_form.interfaces.widget.IWidget."""
        params = self.request.params
        if (self.name not in params and self.name + '-empty-marker' in params):
            return ()
        try:
            value = params.getall(self.name) or default
        except AttributeError:
            value = params.get(self.name, default)
        if value != default:
            values = []
            if not isinstance(value, (tuple, list)):
                # this is here to make any single value a tuple
                value = (value,)
            for val in value:
                values.extend(val.split(self.separator))
            value = tuple(values)
            # do some kind of validation, at least only use existing values
            for token in value:
                if token == self.no_value_token:
                    continue
                try:
                    self.terms.getTermByToken(token)
                except LookupError:
                    return default
        return value

    def json_data(self):
        data = super().json_data()
        data['type'] = 'sequence'
        return data


@implementer(IMultiWidget)
class MultiWidget(Widget):
    # pylint: disable=too-many-instance-attributes
    """None Term based sequence widget base.

    The multi widget is used for ITuple, IList or IDict if no other widget is
    defined.

    Some IList, ITuple or IDict are using another specialized widget if they can
    choose from a collection. e.g. a IList of IChoice. The base class of such
    widget is the ISequenceWidget.

    This widget can handle none collection based sequences and offers add or
    remove values to or from the sequence. Each sequence value get rendered by
    it's own relevant widget. e.g. IList of ITextLine or ITuple of IInt

    Each widget get rendered within a sequence value. This means each internal
    widget will represent one value from the multi widget value. Based on the
    nature of this (sub) widget setup the internal widget do not have a real
    context and can't get bound to it. This makes it impossible to use a
    sequence of collection where the collection needs a context. But that
    should not be a problem since sequence of collection will use the
    SequenceWidget as base.
    """

    allow_adding = True
    allow_removing = True

    widgets = None
    key_widgets = None
    _value = None
    _widgets_updated = False

    _mode = FieldProperty(IWidget['mode'])

    def __init__(self, request):
        super().__init__(request)
        self.widgets = []
        self.key_widgets = []
        self._value = []

    @property
    def is_dict(self):
        """Check field key type"""
        return getattr(self.field, 'key_type', None) is not None

    @property
    def counter_name(self):
        """Counter name getter"""
        return '%s.count' % self.name

    @property
    def counter_marker(self):
        """Counter HTML marker getter"""
        # this get called in render from the template and contains always the
        # right amount of widgets we use.
        return '<input type="hidden" name="%s" value="%d" />' % (
            self.counter_name, len(self.widgets))

    @property
    def mode(self):
        """This gets the subwidgets mode."""
        return self._mode

    @mode.setter
    def mode(self, mode):
        """Subwidgets mode setter"""
        self._mode = mode
        # ensure that we apply the new mode to the widgets
        for w in self.widgets:
            w.mode = mode
        for w in self.key_widgets:
            if w is not None:
                w.mode = mode

    def get_widget(self, idx, prefix=None, type_field="value_type"):
        """Setup widget based on index id with or without value."""
        value_type = getattr(self.field, type_field)
        registry = self.request.registry
        widget = registry.getMultiAdapter((value_type, self.request), IFieldWidget)
        self.set_name(widget, idx, prefix)
        widget.mode = self.mode
        # set widget.form (objectwidget needs this)
        if IFormAware.providedBy(self):
            widget.form = self.form
            alsoProvides(widget, IFormAware)
        widget.update()
        return widget

    def set_name(self, widget, idx, prefix=None):
        """Set widget name based on index position"""
        names = lambda id: [str(n) for n in [id] + [prefix, idx] if n is not None]
        widget.name = '.'.join([str(self.name)] + names(None))
        widget.id = '-'.join([str(self.id)] + names(None))  # pylint: disable=no-member

    def append_adding_widget(self):
        """Simply append a new empty widget with correct (counter) name."""
        # since we start with idx 0 (zero) we can use the len as next idx
        idx = len(self.widgets)
        widget = self.get_widget(idx)
        self.widgets.append(widget)
        if self.is_dict:
            widget = self.get_widget(idx, "key", "key_type")
            self.key_widgets.append(widget)
        else:
            self.key_widgets.append(None)

    def remove_widgets(self, names):
        """
        :param names: list of widget.name to remove from the value
        :return: None
        """
        zipped = list(zip(self.key_widgets, self.widgets))
        self.key_widgets = [k for k, v in zipped if v.name not in names]
        self.widgets = [v for k, v in zipped if v.name not in names]
        if self.is_dict:
            self.value = [(k.value, v.value)
                          for k, v in zip(self.key_widgets, self.widgets)]
        else:
            self.value = [widget.value for widget in self.widgets]

    def apply_value(self, widget, value=NO_VALUE):
        """Validate and apply value to given widget.

        This method gets called on any multi widget value change and is
        responsible for validating the given value and setup an error message.

        This is internal apply value and validation process is needed because
        nothing outside this multi widget does know something about our
        internal sub widgets.
        """
        apply_value_to_widget(self, widget, value)

    def update_widgets(self):  # pylint: disable=too-many-branches
        """Setup internal widgets based on the value_type for each value item.
        """
        old_len = len(self.widgets)
        # Ensure at least min_length widgets are shown
        if (IMinMaxLen.providedBy(self.field) and
                self.mode == INPUT_MODE and self.allow_adding and
                old_len < self.field.min_length):
            old_len = self.field.min_length
        self.widgets = []
        self.key_widgets = []
        keys = set()
        idx = 0
        if self.value:
            if self.is_dict:
                # mainly sorting for testing reasons
                # and dict's have no order!, but
                # XXX: this should be done in the converter... here we get
                #      always strings as keys, sorting an str(int/date) is lame
                #      also, newly added item should be the last...
                try:
                    items = sorted_none(self.value)
                except:  # pylint: disable=bare-except
                    # just in case it's impossible to sort don't fail
                    items = self.value
            else:
                items = zip([None] * len(self.value), self.value)
            for key, v in items:
                widget = self.get_widget(idx)
                self.apply_value(widget, v)
                self.widgets.append(widget)

                if self.is_dict:
                    # This is needed, since sequence widgets (such as for
                    # choices) return lists of values.
                    hash_key = key if not isinstance(key, list) else tuple(key)
                    widget = self.get_widget(idx, "key", "key_type")
                    self.apply_value(widget, key)
                    if hash_key in keys and widget.error is None:
                        error = Invalid('Duplicate key')
                        registry = self.request.registry
                        view = registry.getMultiAdapter((error, self.request, widget,
                                                         widget.field, self.form, self.context),
                                                        IErrorViewSnippet)
                        view.update()
                        widget.error = view
                    self.key_widgets.append(widget)
                    keys.add(hash_key)
                else:
                    # makes the template easier to have this the same length
                    self.key_widgets.append(None)
                idx += 1
        missing = old_len - len(self.widgets)
        if missing > 0:
            # add previous existing new added widgtes
            for _i in range(missing):
                widget = self.get_widget(idx)
                self.widgets.append(widget)
                if self.is_dict:
                    widget = self.get_widget(idx, "key", "key_type")
                    self.key_widgets.append(widget)
                else:
                    self.key_widgets.append(None)
                idx += 1
        self._widgets_updated = True

    def update_allow_add_remove(self):
        """Update the allow_adding/allow_removing attributes
        basing on field constraints and current number of widgets
        """
        if not IMinMaxLen.providedBy(self.field):
            return
        max_length = self.field.max_length
        min_length = self.field.min_length
        num_items = len(self.widgets)
        self.allow_adding = bool((not max_length) or (num_items < max_length))
        self.allow_removing = bool(num_items and (num_items > min_length))

    @property
    def value(self):
        """This invokes updateWidgets on any value change e.g. update/extract."""
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        # ensure that we apply our new values to the widgets
        self.update_widgets()

    def update(self):
        """See z3c.form.interfaces.IWidget."""
        # Ensure that updateWidgets is called.
        super().update()
        if not self._widgets_updated:
            self.update_widgets()

    def extract(self, default=NO_VALUE):
        # This method is responsible to get the widgets value based on the
        # request and nothing else.
        # We have to setup the widgets for extract their values, because we
        # don't know how to do this for every field without the right widgets.
        # Later we will setup the widgets based on this values. This is needed
        # because we probably set a new value in the form for our multi widget
        # which would generate a different set of widgets.
        params = self.request.params
        if params.get(self.counter_name) is None:
            # counter marker not found
            return NO_VALUE
        counter = int(params.get(self.counter_name, 0))
        # extract value for existing widgets
        values = []
        append = values.append
        # extract value for existing widgets
        for idx in range(counter):
            widget = self.get_widget(idx)
            if self.is_dict:
                key_widget = self.get_widget(idx, "key", "key_type")
                append((key_widget.value, widget.value))
            else:
                append(widget.value)
        return values

    def json_data(self):
        data = super().json_data()
        data['widgets'] = [widget.json_data() for widget in self.widgets]
        data['type'] = 'multi'
        return data


def FieldWidget(field, widget):  # pylint: disable=invalid-name
    """Set the field for the widget."""
    widget.field = field
    if not IFieldWidget.providedBy(widget):
        alsoProvides(widget, IFieldWidget)
    # Initial values are set. They can be overridden while updating the widget
    # itself later on.
    widget.name = field.__name__
    widget.id = field.__name__.replace('.', '-')
    widget.label = field.title
    widget.required = field.required
    return widget


@implementer(IWidgetEvent)
class WidgetEvent:
    """Base widget event"""

    def __init__(self, widget):
        self.widget = widget

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.widget)


@implementer_only(IAfterWidgetUpdateEvent)
class AfterWidgetUpdateEvent(WidgetEvent):
    """Widget after update event"""


class WidgetSelector:
    """Widget event selector

    This predicate can be used by subscribers to filter widgets events.
    """

    def __init__(self, ifaces, config):  # pylint: disable=unused-argument
        if not isinstance(ifaces, (list, tuple)):
            ifaces = (ifaces,)
        self.interfaces = ifaces

    def text(self):
        """Widget's selector text"""
        return 'widget_selector = %s' % str(self.interfaces)

    phash = text

    def __call__(self, event):
        for intf in self.interfaces:
            try:
                if intf.providedBy(event.widget):
                    return True
            except (AttributeError, TypeError):
                if isinstance(event.widget, intf):
                    return True
        return False



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.field module

Form field definitions.
"""

from zope.interface import Interface, Invalid, alsoProvides, implementer
from zope.interface.interface import InterfaceClass
from zope.location import locate
from zope.schema import getFieldsInOrder
from zope.schema.interfaces import IField as ISchemaField

from pyams_form.error import MultipleErrors
from pyams_form.interfaces import DISPLAY_MODE, IDataConverter, IField, IFields, \
    IManagerValidator, INPUT_MODE, IValidator
from pyams_form.interfaces.error import IErrorViewSnippet
from pyams_form.interfaces.form import IContextAware, IFieldsForm, IFormAware
from pyams_form.interfaces.widget import IFieldWidget, IWidgets
from pyams_form.util import Manager, SelectionManager, expand_prefix
from pyams_form.widget import AfterWidgetUpdateEvent
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import IDataManager, NO_VALUE
from pyams_utils.registry import get_current_registry


__docformat__ = 'restructuredtext'


def _initkw(keep_readonly=(), omit_readonly=False, **defaults):
    """Init keywords"""
    return keep_readonly, omit_readonly, defaults


class WidgetFactories(dict):
    """Widget factories"""

    def __init__(self):
        super().__init__()
        self.default = None

    def __getitem__(self, key):
        if key not in self and self.default:
            return self.default
        return super().__getitem__(key)

    def get(self, key, default=None):
        if key not in self and self.default:
            return self.default
        return super().get(key, default)


class WidgetFactoryProperty:
    """Widget factory property"""

    def __get__(self, inst, klass):
        if not hasattr(inst, '_widget_factories'):
            inst._widget_factories = WidgetFactories()
        return inst._widget_factories

    def __set__(self, inst, value):
        if not hasattr(inst, '_widget_factories'):
            inst._widget_factories = WidgetFactories()
        inst._widget_factories.default = value


@implementer(IField)
class Field:
    """Field implementation."""

    widget_factory = WidgetFactoryProperty()

    # pylint: disable=too-many-arguments
    def __init__(self, field, name=None, prefix='', mode=None, interface=None,
                 ignore_context=None, show_default=None):
        self.field = field
        if name is None:
            name = field.__name__
        assert name
        self.__name__ = expand_prefix(prefix) + name
        self.prefix = prefix
        self.mode = mode
        if interface is None:
            interface = field.interface
        self.interface = interface
        self.ignore_context = ignore_context
        self.show_default = show_default

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


@implementer(IFields)
class Fields(SelectionManager):
    """Field manager."""

    manager_interface = IFields

    def __init__(self, *args, **kw):  # pylint: disable=too-many-branches
        keep_readonly, omit_readonly, defaults = _initkw(**kw)

        fields = []
        for arg in args:
            if isinstance(arg, InterfaceClass):
                for name, field in getFieldsInOrder(arg):
                    fields.append((name, field, arg))

            elif ISchemaField.providedBy(arg):
                name = arg.__name__
                if not name:
                    raise ValueError("Field has no name")
                fields.append((name, arg, arg.interface))

            elif self.manager_interface.providedBy(arg):
                for form_field in arg.values():
                    fields.append(
                        (form_field.__name__, form_field, form_field.interface))

            elif isinstance(arg, Field):
                fields.append((arg.__name__, arg, arg.interface))

            else:
                raise TypeError("Unrecognized argument type", arg)

        super().__init__()
        for name, field, iface in fields:
            if isinstance(field, Field):
                form_field = field
            else:
                if field.readonly:
                    if omit_readonly and (name not in keep_readonly):
                        continue
                custom_defaults = defaults.copy()
                if iface is not None:
                    custom_defaults['interface'] = iface
                form_field = Field(field, **custom_defaults)
                name = form_field.__name__

            if name in self:
                raise ValueError("Duplicate name", name)

            self[name] = form_field

    def select(self, *names, **kwargs):  # pylint: disable=arguments-differ
        """See interfaces.IFields"""
        prefix = kwargs.pop('prefix', None)
        interface = kwargs.pop('interface', None)
        assert len(kwargs) == 0
        if prefix:
            names = [expand_prefix(prefix) + name for name in names]
        mapping = self
        if interface is not None:
            mapping = {field.field.__name__: field for field in self.values()
                       if field.field.interface is interface}
        return self.__class__(*[mapping[name] for name in names])

    def omit(self, *names, **kwargs):  # pylint: disable=arguments-differ
        """See interfaces.IFields"""
        prefix = kwargs.pop('prefix', None)
        interface = kwargs.pop('interface', None)
        assert len(kwargs) == 0
        if prefix:
            names = [expand_prefix(prefix) + name for name in names]
        return self.__class__(
            *[field for name, field in self.items()
              if not ((name in names and interface is None) or
                      (field.field.interface is interface and
                       field.field.__name__ in names))])


@adapter_config(required=(IFieldsForm, IFormLayer, Interface),
                provides=IWidgets)
class FieldWidgets(Manager):
    """Widget manager for IFieldWidget."""

    prefix = 'widgets.'
    mode = INPUT_MODE
    errors = ()
    has_required_fields = False
    ignore_context = False
    ignore_request = False
    ignore_readonly = False
    ignore_required_on_extract = False
    set_errors = True

    def __init__(self, form, request, content):
        super().__init__()
        self.form = form
        self.request = request
        self.content = content

    def validate(self, data):
        """Validate widgets fields"""
        fields = self.form.fields.values()

        # Step 1: Collect the data for the various schemas
        schema_data = {}
        for field in fields:
            schema = field.interface
            if schema is None:
                continue

            field_data = schema_data.setdefault(schema, {})
            if field.__name__ in data:
                field_data[field.field.__name__] = data[field.__name__]

        # Step 2: Validate the individual schemas and collect errors
        errors = ()
        content = self.content
        if self.ignore_context:
            content = None
        registry = self.request.registry
        for schema, field_data in schema_data.items():
            validator = registry.getMultiAdapter((content, self.request, self.form, schema, self),
                                                 IManagerValidator)
            errors += validator.validate(field_data)

        return errors

    def update(self):
        """See interfaces.widget.IWidgets"""
        # Create a unique prefix.
        prefix = expand_prefix(self.form.prefix) + expand_prefix(self.prefix)
        # Walk through each field, making a widget out of it.
        data = {}
        data.update(self)
        registry = self.request.registry
        for field in self.form.fields.values():
            # Step 0. Determine whether the context should be ignored.
            ignore_context = self.ignore_context
            if field.ignore_context is not None:
                ignore_context = field.ignore_context
            # Step 1: Determine the mode of the widget.
            mode = self.mode
            if field.mode is not None:
                mode = field.mode
            elif field.field.readonly and not self.ignore_readonly:
                mode = DISPLAY_MODE
            elif not ignore_context:
                # If we do not have enough permissions to write to the
                # attribute, then switch to display mode.
                dman = registry.getMultiAdapter((self.content, field.field), IDataManager)
                if not dman.can_write():
                    mode = DISPLAY_MODE
            # Step 2: Get the widget for the given field.
            short_name = field.__name__
            new_widget = True
            if short_name in self:
                # reuse existing widget
                widget = data[short_name]
                new_widget = False
            elif field.widget_factory.get(mode) is not None:
                factory = field.widget_factory.get(mode)
                widget = factory(field.field, self.request)
            else:
                widget = registry.getMultiAdapter((field.field, self.request), IFieldWidget)
            # Step 3: Set the prefix for the widget
            widget.name = prefix + short_name
            widget.id = (prefix + short_name).replace('.', '-')
            # Step 4: Set the context
            widget.context = self.content
            # Step 5: Set the form
            widget.form = self.form
            # Optimization: Set both interfaces here, rather in step 4 and 5:
            # ``alsoProvides`` is quite slow
            alsoProvides(widget, IContextAware, IFormAware)
            # Step 6: Set some variables
            widget.ignore_context = ignore_context
            widget.ignore_request = self.ignore_request
            if field.show_default is not None:
                widget.show_default = field.show_default
            # Step 7: Set the mode of the widget
            widget.mode = mode
            # Step 8: Update the widget
            widget.update()
            get_current_registry().notify(AfterWidgetUpdateEvent(widget))
            # Step 9: Add the widget to the manager
            if widget.required:
                self.has_required_fields = True
            if new_widget:
                data[short_name] = widget
                locate(widget, self, short_name)
        self.create_according_to_list(data, self.form.fields.keys())

    def _extract(self, return_raw=False):
        data = {}
        errors = ()
        registry = self.request.registry
        for name, widget in self.items():
            if widget.mode == DISPLAY_MODE:
                continue
            value = widget.field.missing_value
            try:
                widget.set_errors = self.set_errors
                raw = widget.extract()
                if raw is not NO_VALUE:
                    # pylint: disable=assignment-from-no-return
                    value = IDataConverter(widget).to_field_value(raw)
                widget.ignore_required_on_validation = self.ignore_required_on_extract
                registry.getMultiAdapter((self.content, self.request, self.form,
                                          getattr(widget, 'field', None), widget),
                                         IValidator).validate(value)
            except (Invalid, ValueError, MultipleErrors) as error:
                view = registry.getMultiAdapter((error, self.request, widget, widget.field,
                                                 self.form, self.content),
                                                IErrorViewSnippet)
                view.update()
                if self.set_errors:
                    widget.error = view
                errors += (view,)
            else:
                name = widget.__name__
                if return_raw:
                    data[name] = raw
                else:
                    data[name] = value
        for error in self.validate(data):
            view = registry.getMultiAdapter((error, self.request, None, None,
                                             self.form, self.content),
                                            IErrorViewSnippet)
            view.update()
            errors += (view,)
        if self.set_errors:
            self.errors = errors
        return data, errors

    def extract(self):
        """See interfaces.IWidgets"""
        return self._extract(return_raw=False)

    def extract_raw(self):
        """See interfaces.IWidgets"""
        return self._extract(return_raw=True)

    def copy(self):
        """See interfaces.ISelectionManager"""
        clone = self.__class__(self.form, self.request, self.content)
        super(self.__class__, clone).update(self)
        return clone



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.interfaces.error module

This module defines error-related interfaces.
"""

from zope.interface import Attribute, Interface
from zope.schema import Field

from pyams_layer.interfaces import IFormLayer
from pyams_template.template import template_config


__docformat__ = 'restructuredtext'

from pyams_form import _


@template_config(layer=IFormLayer,
                 template='templates/error.pt')
class IErrorViewSnippet(Interface):
    """A view providing a view for an error"""

    widget = Field(title=_("Widget"),
                   description=_("The widget that the view is on"),
                   required=True)

    error = Field(title=_('Error'),
                  description=_('Error the view is for'),
                  required=True)

    def update(self):
        """Update view"""

    def render(self):
        """Render view"""


class IMultipleErrors(Interface):
    """An error that contains many status"""

    errors = Attribute("List of status")


class IAJAXErrorsRenderer(Interface):
    """AJAX error manager"""

    def render(self, errors):
        """Render error message in AJAX"""
